from typing import List, Optional, Union

import torch
from loguru import logger


def load_model_and_tokenizer(
    model_name_or_path: str,
    tokenizer_name_or_path: Optional[str] = None,
    device: Optional[torch.device] = None,
    dtype: Optional[Union[str, torch.dtype]] = None,
    device_map: Optional[Union[str, List[int]]] = None,
    **kwargs
):
    """Load a model and tokenizer from HuggingFace."""
    import os

    import huggingface_hub
    from accelerate import init_empty_weights, load_checkpoint_and_dispatch
    from transformers import (
        AutoConfig,
        AutoModelForCausalLM,
        AutoTokenizer,
        GPTNeoXForCausalLM,
    )

    from ...helper import auto_dtype_and_device

    dtype, device = auto_dtype_and_device(dtype, device)
    revision = kwargs.pop('revision', 'step143000')

    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_name_or_path or model_name_or_path,
        revision=revision,
    )

    tokenizer.padding_side = 'left'
    tokenizer.pad_token = tokenizer.eos_token
    tokenizer.pad_token_id = tokenizer.eos_token_id

    if device_map:
        if not os.path.exists(model_name_or_path):
            model_path = huggingface_hub.snapshot_download(model_name_or_path)
        else:
            model_path = model_name_or_path

        config = AutoConfig.from_pretrained(model_name_or_path, trust_remote_code=True)

        with init_empty_weights():
            model = AutoModelForCausalLM.from_config(
                config, torch_dtype=dtype, trust_remote_code=True
            )
            # make sure token embedding weights are still tied if needed
            model.tie_weights()

            model = load_checkpoint_and_dispatch(
                model,
                model_path,
                device_map=device_map,
                dtype=dtype,
                no_split_module_classes=["GPTNeoXLayer"],
            )
    else:

        model = GPTNeoXForCausalLM.from_pretrained(
            model_name_or_path,
            revision=revision,
            torch_dtype=dtype,
            **kwargs,
        )
        model.to(device)

    return model, tokenizer



from typing import TYPE_CHECKING, Optional, Union

from accelerate import init_empty_weights, load_checkpoint_and_dispatch
from transformers import AutoConfig, AutoTokenizer, T5ForConditionalGeneration

if TYPE_CHECKING:
    import torch

from loguru import logger


def load_model_and_tokenizer(
    model_name_or_path: str,
    tokenizer_name_or_path: Optional[str] = None,
    device: Optional[torch.device] = None,
    dtype: Optional[Union[str, torch.dtype]] = None,
    **kwargs
):
    """Load a model and tokenizer from HuggingFace."""

    from ...helper import auto_dtype_and_device

    dtype, device = auto_dtype_and_device(dtype, device)

    load_in_8bit = True if (str(dtype) == 'torch.int8') else False

    model = T5ForConditionalGeneration.from_pretrained(
        model_name_or_path,
        load_in_8bit=load_in_8bit,
        torch_dtype=dtype,
        # device_map="auto",
        **kwargs
    )
    model.to(device)

    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_name_or_path or model_name_or_path
    )

    return model, tokenizer



from typing import List, Optional, Union

import torch
from transformers import StoppingCriteria, StoppingCriteriaList

from open_gpt.models.modeling import BaseModel


class StopOnTokens(StoppingCriteria):
    def __call__(
        self, input_ids: torch.LongTensor, scores: torch.FloatTensor, **kwargs
    ) -> bool:
        stop_ids = [50278, 50279, 50277, 1, 0]
        for stop_id in stop_ids:
            if input_ids[0][-1] == stop_id:
                return True
        return False


class StableLMModel(BaseModel):

    no_split_module_classes = ["GPTNeoXLayer"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def generate(self, prompts: Union[str, List[str]], **kwargs):
        """Generate text from the given prompt."""
        return super().generate(
            prompts, stopping_criteria=StoppingCriteriaList([StopOnTokens()]), **kwargs
        )



"""The model hub is a central repository of models that can be used to load pretrained models."""


def load_model_from_hub(model_name: str, **kwargs):
    """Load a model from the model hub.

    :param model_name: The name of the model to load.
    :param kwargs: Additional arguments to pass to the model.
    :return: The model.
    """
    ...



from typing import TYPE_CHECKING, List, Optional, Union

import torch
from torch import nn

from ..helper import auto_dtype_and_device

if TYPE_CHECKING:
    from transformers import AutoModelForCausalLM, AutoTokenizer


class BaseModel(nn.Module):

    model: 'AutoModelForCausalLM'
    tokenizer: 'AutoTokenizer'
    no_split_module_classes: List[str] = None

    def __init__(
        self,
        model_name_or_path: str,
        tokenizer_name_or_path: Optional[str] = None,
        dtype: Optional[Union[str, torch.dtype]] = None,
        device: Optional[torch.device] = None,
        device_map: Optional[Union[str, List[int]]] = None,
        **kwargs,
    ):
        """Create a model of the given name."""

        super().__init__()

        self._dtype, self._device = auto_dtype_and_device(dtype, device)

        self._device_map = device_map

        self.load_model_and_transforms(
            model_name_or_path, tokenizer_name_or_path=tokenizer_name_or_path
        )

    def load_model_and_transforms(
        self, model_name_or_path: str, tokenizer_name_or_path: Optional[str] = None
    ):
        from .loading import load_model_and_tokenizer

        self.model, self.tokenizer = load_model_and_tokenizer(
            model_name_or_path,
            tokenizer_name_or_path=tokenizer_name_or_path,
            dtype=self._dtype,
            device=self._device,
            device_map=self._device_map,
            no_split_module_classes=self.no_split_module_classes,
        )

        self.model.eval()

    def generate(self, prompts: Union[str, List[str]], **kwargs):
        """Generate text from the given prompt."""

        inputs = self.tokenizer(
            [prompts] if isinstance(prompts, str) else prompts,
            padding=True,
            return_tensors="pt",
        )

        # Move inputs to the correct device
        for k, v in inputs.items():
            if isinstance(v, torch.Tensor):
                inputs[k] = v.to(self._device)

        with torch.inference_mode():
            outputs = self.model.generate(**inputs, **kwargs)

            texts_outs = []
            for _, generated_sequence in enumerate(outputs):
                generated_sequence = generated_sequence.tolist()
                prompt = prompts[_] if isinstance(prompts, list) else prompts
                prompt_len = len(prompt)

                text = self.tokenizer.decode(
                    generated_sequence,
                    clean_up_tokenization_spaces=True,
                    skip_special_tokens=True,
                )

                text = text[prompt_len:] if text[:prompt_len] == prompt else text
                texts_outs.append(text)

            return texts_outs



from typing import List, Optional, Union

import torch
from loguru import logger

from open_gpt.helper import auto_dtype_and_device


def load_model_and_tokenizer(
    model_name_or_path: str,
    tokenizer_name_or_path: Optional[str] = None,
    device: Optional[torch.device] = None,
    dtype: Optional[Union[str, torch.dtype]] = None,
    device_map: Optional[Union[str, List[int]]] = None,
    no_split_module_classes: Optional[List[str]] = None,
    **kwargs,
):
    """Load a model and tokenizer from HuggingFace."""
    import os

    from transformers import AutoConfig, AutoModelForCausalLM, AutoTokenizer

    dtype, device = auto_dtype_and_device(dtype, device)

    tokenizer = AutoTokenizer.from_pretrained(
        tokenizer_name_or_path or model_name_or_path, trust_remote_code=True
    )

    if tokenizer.pad_token is None:
        # Issue: GPT models don't have a pad token
        # tokenizer.add_special_tokens({"pad_token": "<PAD>"})
        tokenizer.pad_token = tokenizer.eos_token
        tokenizer.pad_token_id = tokenizer.eos_token_id

    # For generation padding tokens should be on the left
    tokenizer.padding_side = "left"

    if device_map:
        import huggingface_hub
        from accelerate import init_empty_weights, load_checkpoint_and_dispatch

        if not os.path.exists(model_name_or_path):
            model_path = huggingface_hub.snapshot_download(model_name_or_path)
        else:
            model_path = model_name_or_path

        config = AutoConfig.from_pretrained(model_name_or_path, trust_remote_code=True)
        with init_empty_weights():
            model = AutoModelForCausalLM.from_config(
                config, torch_dtype=dtype, trust_remote_code=True
            )
            # make sure token embedding weights are still tied if needed
            model.tie_weights()

            model = load_checkpoint_and_dispatch(
                model,
                model_path,
                device_map=device_map,
                no_split_module_classes=no_split_module_classes,
                dtype=dtype,
            )
    else:
        model = AutoModelForCausalLM.from_pretrained(
            model_name_or_path, torch_dtype=dtype, trust_remote_code=True
        )
        model.to(device)

    return model, tokenizer


# def create_model_and_transforms(
#     model_name: str,
#     device: Optional[Union[str, torch.device]] = None,
#     precision: Optional[str] = None,
#     **kwargs,
# ):
#     """Create a model of the given name.
#
#     :param model_name: The name of the model to create.
#     :param device: The device to create the model on.
#     :param precision: The precision to use for the model.
#     :param kwargs: Additional arguments to pass to the model.
#     :return: The model.
#     """
#
#     dtype, device = auto_dtype_and_device(precision, device)
#
#     # TODO: Add support for loading config based on model name
#     model_config = {}
#
#     logger.info(
#         f'Loading "{model_name}" with precision: `{dtype}` on device: `{device}`'
#     )
#
#     if model_name.startswith('openflamingo/OpenFlamingo'):
#         from .flamingo.loading import load_model_and_transforms
#
#         model_config = {
#             'vision_model_name_or_path': 'ViT-L-14::openai',
#             'lang_model_name_or_path': 'llama_7B',
#             'tokenizer_name_or_path': 'llama_7B',
#         }
#         return load_model_and_transforms(
#             model_name, device=device, dtype=dtype, **model_config
#         )
#     elif model_name.startswith('facebook/llama'):
#         from .llama.loading import load_model_and_tokenizer
#
#         model_config = {
#             'model_name_or_path': 'llama_7B',
#             'tokenizer_name_or_path': 'llama_7B',
#         }
#         return load_model_and_tokenizer(
#             model_name, device=device, dtype=dtype, **model_config
#         )
#     elif model_name.startswith('google/flan'):
#         from .flan.loading import load_model_and_tokenizer
#
#         return load_model_and_tokenizer(
#             model_name, device=device, dtype=dtype, **model_config
#         )
#     elif model_name.startswith('EleutherAI/pythia'):
#         from .pythia.loading import load_model_and_tokenizer
#
#         return load_model_and_tokenizer(
#             model_name, device=device, dtype=dtype, **model_config
#         )
#     elif model_name.startswith('stabilityai/stablelm'):
#         from .stablelm.loading import load_model_and_tokenizer
#
#         return load_model_and_tokenizer(
#             model_name, device=device, dtype=dtype, **model_config
#         )
#     elif model_name.startswith('fnlp/moss-moon'):
#         from .moss.loading import load_model_and_tokenizer
#
#         return load_model_and_tokenizer(
#             model_name, device=device, dtype=dtype, **model_config, **kwargs
#         )
#     else:
#         raise ValueError(f'Unknown model name: {model_name}')



from typing import List, Optional, Union

import torch

from open_gpt.models.llama.modeling import LlamaModel


class VicunaModel(LlamaModel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def load_model_and_transforms(
        self, model_name_or_path: str, tokenizer_name_or_path: Optional[str] = None
    ):

        from .loading import load_model_and_tokenizer

        self.model, self.tokenizer = load_model_and_tokenizer(
            model_name_or_path,
            tokenizer_name_or_path=tokenizer_name_or_path,
            dtype=self._dtype,
            device=self._device,
            device_map=self._device_map,
            no_split_module_classes=self.no_split_module_classes,
        )

        self.model.eval()



"""The model adapter for OpenGPT models.

The model adapter is a wrapper around the model that provides a light-weight way to
enrich the model with additional functionality.
"""
import torch
from torch import nn


class ModelAdapter(nn.Module):
    def __init__(self, base_model: nn.Module, adapter_configs: dict, **kwargs):
        super().__init__()
        self._base_model = base_model
        self._adapter_configs = adapter_configs

    def forward(self):
        ...



from typing import Optional

from open_gpt.models.modeling import BaseModel


class PythiaModel(BaseModel):
    no_split_module_classes = ["GPTNeoXLayer"]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
